from pathlib import Path
from typing import ClassVar

from arrest.openapi._base import TemplateBase


class InitTemplate(TemplateBase):
    TEMPLATE_FILEPATH: ClassVar[str] = "__init__.jinja2"

    def __init__(self, *, destination_path: Path | str) -> None:
        super().__init__(source=self.TEMPLATE_FILEPATH, destination_path=destination_path)



from pathlib import Path
from typing import ClassVar, Optional

from pydantic import BaseModel

from arrest.openapi._base import TemplateBase


class ServiceSchema(BaseModel):
    service_id: str
    name: str
    url: str
    description: Optional[str] = None
    resources: list[str]


class ServiceParams(BaseModel):
    resource_imports: set[str]
    services: list[ServiceSchema]


class ServiceTemplate(TemplateBase):
    TEMPLATE_FILEPATH: ClassVar[str] = "services.jinja2"

    def __init__(self, services: list[ServiceSchema], destination_path: Path | str) -> None:
        resource_imports = set()
        for service in services:
            resource_imports.update(service.resources)

        super().__init__(
            source=self.TEMPLATE_FILEPATH,
            params=ServiceParams(resource_imports=resource_imports, services=services),
            destination_path=destination_path,
        )



import itertools
from functools import partial
from typing import Any, Optional

import httpx
from typing_extensions import Unpack

from arrest._config import HttpxClientInputs
from arrest.exceptions import ResourceNotFound
from arrest.http import Methods
from arrest.resource import Resource


class Service:
    def __init__(
        self,
        name: str,
        url: str,
        description: Optional[str] = None,
        resources: Optional[list[Resource]] = [],
        client: Optional[httpx.AsyncClient] = None,
        **kwargs: Unpack[HttpxClientInputs],
    ) -> None:
        """
        A python class to define a service.
        Contains a base url for the main HTTP service.
        Resources can be added to a service.

        Parameters:
            name:
                Name of the service
            url:
                Base url of the service
            resources:
                A list of resources provided by the service
            client:
                An httpx.AsyncClient instance
            kwargs:
                Additional httpx.AsyncClient parameters. [see more](api.md/#httpx-client-arguments)
        """

        self.name = name
        self.url = url
        self.description = description
        self.resources: dict[str, Resource] = {}
        for resource in resources:
            self.add_resource(resource, client=client, **kwargs)

        self.get = partial(self.request, method=Methods.GET)
        self.post = partial(self.request, method=Methods.POST)
        self.put = partial(self.request, method=Methods.PUT)
        self.patch = partial(self.request, method=Methods.PATCH)
        self.delete = partial(self.request, method=Methods.DELETE)
        self.head = partial(self.request, method=Methods.HEAD)
        self.options = partial(self.request, method=Methods.OPTIONS)

    def add_resource(
        self,
        resource: Resource,
        client: Optional[httpx.AsyncClient] = None,
        **kwargs: Unpack[HttpxClientInputs],
    ) -> None:
        """
        Add a new resource to the service

        Parameters:
            resource:
                The new resource instance
        """
        resource.base_url = self.url
        resource.initialize_handlers(base_url=self.url)
        resource.initialize_httpx(client=client, **kwargs)
        self.resources[resource.name] = resource
        setattr(self, resource.name, resource)

    async def request(self, path: str, method: Methods, **kwargs):
        """
        Helper function to make a request directly
        from the service level

        Parameters:
            path:
                Requested path needs to have the following syntax:
                `/{resource_route}/{handler_route}`
            method:
                Requested HTTP Method
        """
        parts = path.strip("/").split("/")
        resource, suffix = parts[0], "/".join(parts[1:])

        if resource not in self.resources:
            raise ResourceNotFound(message=f"resource {resource} not found")
        return await self.resources[resource].request(path=f"/{suffix}", method=method, **kwargs)

    def __getattr__(self, key: str) -> Resource | Any:  # pragma: no cover
        if hasattr(self, key):
            return getattr(self, key)
        return self.resources[key]

    def __dir__(self):  # pragma: no cover
        return list(itertools.chain(dir(super()), self.resources))



import re
from typing import Any

from arrest.openapi.spec import Reference


def get_ref_schema(reference: Reference | Any) -> str | None:
    if ref := getattr(reference, "ref", None):
        return ref.split("/")[-1]


def sanitize_name(name: str) -> str:
    name = name.lower().replace(" ", "_")
    return re.sub("[^A-Za-z0-9]+", "_", name)



from argparse import ArgumentParser

arg_parser = ArgumentParser(
    usage="\n arrest [options]",
    description="generate arrest services and resources from various definitions",
    add_help=True,
)

arg_parser.add_argument(
    "-o", "--output", default=None, help="output directory for generated files (default: ./api)"
)

arg_parser.add_argument(
    "--pydantic",
    choices=["v1", "v2"],
    default="v1",
    help="pydantic version to generate the schema definitions",
)
arg_parser.add_argument("-u", "--url", default=None, help="HTTP or file url for the openapi schema")
arg_parser.add_argument("-d", "--dir", default=None, help="Directory containing the files")



from functools import cached_property, lru_cache
from pathlib import Path
from typing import Optional

from jinja2 import Environment, FileSystemLoader, Template
from pydantic import BaseModel

from arrest._config import PYDANTIC_V2
from arrest.openapi._config import TEMPLATE_DIR


@lru_cache()
def get_template(template_filepath: Path) -> Template:
    template_dir = Path(__file__).parents[0] / TEMPLATE_DIR
    loader = FileSystemLoader(str(template_dir / template_filepath.parent))
    environment = Environment(loader=loader)
    return environment.get_template(template_filepath.name)


class TemplateBase:
    def __init__(
        self, *, source: str, params: Optional[BaseModel] = None, destination_path: Path | str
    ) -> None:
        self.source = Path(source)
        self.params = params
        self.destination_path = destination_path

    @cached_property
    def template(self):
        return get_template(template_filepath=self.source)

    def render(self) -> str:
        if self.params:
            kwargs: dict = self.params.model_dump() if PYDANTIC_V2 else self.params.dict()
            return self.template.render(**kwargs)
        else:
            return self.template.render()

    def render_and_save(self) -> None:
        content = self.render()
        output_filename = f"{self.source.stem}.py"
        with open(Path(self.destination_path / output_filename), "w", encoding="utf-8") as file:
            file.write(content)



import difflib
import re
from typing import Callable, NamedTuple, Pattern, Type

from pydantic import AnyUrl, BaseModel

from arrest.converters import replace_params
from arrest.exceptions import ConversionError
from arrest.http import Methods
from arrest.logging import logger


class HandlerKey(NamedTuple):
    method: Methods
    route: str


class ResourceHandler(BaseModel):
    """
    A pydantic class defining a resource handler

    Parameters:
        method (Methods):
            HTTP Method for the handler
        route (str):
            Unique path to the handler from its parent resource
        request (Type[BaseModel], optional):
            Pydantic type to validate the request with
        response (Type[BaseModel], optional):
            Pydantic type to deserialize the HTTP response
        callback (Callable, optional):
            A callable (sync or async) to execute with the HTTP
            response
    """

    method: Methods
    route: str
    request: Type[BaseModel] | None = None
    response: Type[BaseModel] | None = None
    callback: Callable | None = None
    path_format: str | None = None
    path_regex: Pattern | None = None
    param_types: dict[str, type] | None = None

    def parse_path(self, method: Methods, path: str | AnyUrl, **kwargs) -> str | AnyUrl | None:
        if method != self.method:
            return None

        if kwargs:
            return self.__resolve_path_param(path, **kwargs)

        return self.__parse_exact_path(path)

    def __parse_exact_path(self, path: str | AnyUrl) -> str | AnyUrl | None:
        if self.path_regex.fullmatch(path):
            return path

    def __resolve_path_param(self, path: str | AnyUrl, **kwargs) -> str | AnyUrl | None:
        params = self.__extract_path_params(path)
        if params is None:
            return None

        params |= kwargs

        try:
            parsed_path, remaining_params = replace_params(
                path=self.path_format,
                path_params=params,
                param_types=self.param_types,
            )
        except ConversionError as exc:
            logger.warning(str(exc), exc_info=True)
            return None
        if remaining_params:
            return None

        return self.__parse_exact_path(parsed_path)

    def __extract_path_params(self, path: str | AnyUrl) -> dict | None:
        differ = difflib.Differ()
        diff = list(differ.compare(self.path_format.split("/"), path.split("/")))

        left = [delta.lstrip("- ") for delta in diff if delta.startswith("- ")]
        right = [delta.lstrip("+ ") for delta in diff if delta.startswith("+ ")]

        params = dict(zip(left, right))
        for key in list(params):
            if not (match := re.match("{(.*)}", key)):
                return None

            params[match.group(1)] = params.pop(key)

        return {k: v for k, v in params.items() if v}



import os

DEFAULT_TIMEOUT = int(os.getenv("ARREST_DEFAULT_TIMEOUT", "60"))  # sec
MAX_RETRIES = int(os.getenv("ARREST_MAX_RETRIES", "3"))

OPENAPI_SCHEMA_FILENAME = "models.py"
SERVICE_FILENAME = "service.py"
RESOURCE_FILENAME = "resources.py"
OPENAPI_DIRECTORY = "api"



from typing import Any, Callable, Mapping, Optional, TypedDict, Union

from httpx import AsyncBaseTransport, Limits, _types
from pydantic.version import VERSION as PYDANTIC_VERSION

PYDANTIC_V2 = PYDANTIC_VERSION.startswith("2")


class HttpxClientInputs(TypedDict, total=False):
    """
    a typeddict to check for all the necessary fields
    for building an `httpx.AsyncClient` instance
    """

    auth: Optional[_types.AuthTypes]
    params: Optional[_types.QueryParamTypes]
    headers: Optional[_types.HeaderTypes]
    cookies: Optional[_types.CookieTypes]
    verify: Optional[_types.VerifyTypes]
    cert: Optional[_types.CertTypes]
    http2: Optional[bool]
    proxies: Optional[_types.ProxiesTypes]
    mounts: Optional[Mapping[str, AsyncBaseTransport]]
    timeout: Optional[_types.TimeoutTypes]
    follow_redirects: Optional[bool]
    limits: Optional[Limits]
    max_redirects: Optional[int]
    event_hooks: Optional[Mapping[str, list[Callable[..., Any]]]]
    transport: Optional[AsyncBaseTransport]
    trust_env: Optional[bool]
    default_encoding: Union[str, Callable[[bytes], str]]



"""
https://github.com/encode/starlette/blob/master/starlette/routing.py
"""

import math
import re
import uuid
from typing import Any, ClassVar, Generic, Mapping, Pattern, TypeVar
from uuid import UUID

from arrest.exceptions import ConversionError

T = TypeVar("T")

# Match parameters in URL paths, eg. '{param}', and '{param:int}'
PARAM_REGEX = re.compile("{([a-zA-Z_][a-zA-Z0-9_]*)(:[a-zA-Z_][a-zA-Z0-9_]*)?}")


class Converter(Generic[T]):
    regex: ClassVar[str] = ""

    def to_str(self, value: T) -> str:
        raise NotImplementedError()


class IntegerConverter(Converter[int]):
    regex = "[0-9]+"

    def to_str(self, value: Any | int) -> str:
        value = int(value)

        assert value >= 0, "Negative integers are not supported"
        return str(value)


class FloatConverter(Converter[float]):
    regex = r"[0-9]+(\.[0-9]+)?"

    def to_str(self, value: Any | float) -> str:
        value = float(value)

        assert value >= 0.0, "Negative floats are not supported"
        assert not math.isnan(value), "NaN values are not supported"
        assert not math.isinf(value), "Infinite values are not supported"
        return ("%0.20f" % value).rstrip("0").rstrip(".")


class StrConverter(Converter[str]):
    regex = "[^/]+"

    def to_str(self, value: Any | str) -> str:
        value = str(value)

        assert "/" not in value, "May not contain path separators"
        assert value, "Must not be empty"
        return value


class UUIDConverter(Converter[uuid.UUID]):
    regex = "[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}"

    def to_str(self, value: Any | UUID) -> str:
        value = UUID(str(value))
        return str(value)


CONVERTER_REGEX: Mapping[str, Converter[Any]] = {
    "str": StrConverter(),
    "int": IntegerConverter(),
    "float": FloatConverter(),
    "uuid": UUIDConverter(),
}


def compile_path(path: str) -> tuple[Pattern[str], str, dict[str, type]]:
    """
    Given a path string, like: "/{username:str}",


    Returns
    -------

    regex : "/(?P<username>[^/]+)"
    format : "/{username}"
    param_types : dict[username: Converter[str]()]
    """
    path_regex = "^"
    path_format = ""

    idx = 0
    parsed_path_params: dict[str, type] = {}

    for match in PARAM_REGEX.finditer(path):
        param_name, converter_type = match.groups("str")
        converter_type = converter_type.lstrip(":").casefold()

        assert (
            converter_type in CONVERTER_REGEX
        ), f"Invalid converter specified, available converters {CONVERTER_REGEX.keys()}"

        converter = CONVERTER_REGEX[converter_type]

        path_regex += re.escape(path[idx : match.start()])
        path_regex += f"(?P<{param_name}>{converter.regex})"

        path_format += path[idx : match.start()]
        path_format += "{%s}" % param_name

        if param_name in parsed_path_params:
            raise ValueError(f"Duplicate param {param_name} at path {path}")
        parsed_path_params[param_name] = converter

        idx = match.end()

    path_regex += re.escape(path[idx:]) + "$"
    path_format += path[idx:]

    pattern = re.compile(path_regex)
    return pattern, path_format, parsed_path_params


def replace_params(
    path: str,
    path_params: dict[str, Any],
    param_types: dict[str, Converter[Any]] | None,
) -> tuple[str, dict[str, str]]:
    for key, value in list(path_params.items()):
        try:
            if "{" + key + "}" in path:
                if not param_types:
                    strval = str(value)
                else:
                    strval = param_types[key].to_str(value)
                path = path.replace("{" + key + "}", strval)
                path_params.pop(key)
        except (TypeError, ValueError) as exc:
            raise ConversionError(*exc.args) from exc

    return path, path_params


def get_converter(key: str) -> Converter[Any]:
    return CONVERTER_REGEX[key]


def add_converter(converter: Converter[Any], key: str) -> None:
    CONVERTER_REGEX[key] = converter



"""
OpenAPI Generator module

Note: only supports >=v3.0 specifications

Dependencies:
    - datamodel-code-generator
"""
import io
import itertools
import json
import os
import sys
from pathlib import Path
from typing import IO, Generator, Optional

import backoff
import httpx
import yaml

from arrest.defaults import MAX_RETRIES, OPENAPI_DIRECTORY, OPENAPI_SCHEMA_FILENAME
from arrest.exceptions import ArrestError
from arrest.http import Methods
from arrest.logging import logger

try:
    from datamodel_code_generator import DataModelType, InputFileType, OpenAPIScope, generate
except ImportError:
    sys.exit(1)

from arrest.openapi._config import Format
from arrest.openapi.init_template import InitTemplate
from arrest.openapi.resource_template import HandlerSchema, ResourceSchema, ResourceTemplate
from arrest.openapi.service_template import ServiceSchema, ServiceTemplate
from arrest.openapi.spec import OpenAPI, Operation, PathItem, Reference, Server
from arrest.openapi.utils import get_ref_schema, sanitize_name


class OpenAPIGenerator:
    def __init__(
        self,
        *,
        url: str,
        output_path: str,
        dir_name: Optional[str] = None,
        use_pydantic_v2: Optional[bool] = False,
    ) -> None:
        """
        class for generating Arrest services, resources and schema
        components from OpenAPI specification (>= v3.0)
        Generates three files:

        1. models.py (contains OpenAPI Schema component definitions)
        2. resources.py (Arrest Resources based on the path items)
        3. services.py (Arrest Services using the resources)


        Parameters:
            url:
                an HTTP url or full path to the OpenAPI specification (json or yaml)
            output_path:
                path where the generated files will be saved
            dir_name:
                (optional) specify the folder name containing the files

        """
        self.url: str = url
        self.output_path: str = output_path
        self.dir_name: str = dir_name
        self.use_pydantic_v2 = use_pydantic_v2

    @backoff.on_exception(
        backoff.expo,
        (httpx.HTTPError, httpx.TimeoutException),
        max_tries=MAX_RETRIES,
        jitter=backoff.full_jitter,
    )
    def download_openapi_spec(self) -> bytes:
        if self.url.startswith("http"):
            response = httpx.get(self.url)
            response.raise_for_status()
            return response.read()

        else:
            with open(self.url, "rb") as file:
                return file.read()

    def generate_schema(self, fmt: Optional[Format] = None):
        """Generates the boilerplate files against an OpenAPI Spec

        Parameters:
            fmt (Optional[Format], optional): specification format [json, yaml, yml]

        Raises:
            ArrestError
        """
        openapi_bytes = self.download_openapi_spec()
        fmt = fmt if fmt else self.url.split(".")[-1]
        openapi: OpenAPI = self.parse_openapi(fmt=fmt, data=io.BytesIO(openapi_bytes))

        output_path = Path(self.output_path)
        if not output_path.exists():
            raise ArrestError("output path does not exist")

        service_name = self.dir_name or self.get_service_name(openapi)
        output_path = output_path / service_name
        schema_path = output_path / OPENAPI_SCHEMA_FILENAME

        Path.mkdir(output_path, exist_ok=True)

        self.generate_component_schema(input_bytes=openapi_bytes, schema_path=schema_path)
        resources = self.generate_resource_file(
            openapi=openapi, schema_path=schema_path, resource_path=output_path
        )
        self.generate_service_file(openapi=openapi, service_path=output_path, resources=resources)
        InitTemplate(destination_path=output_path).render_and_save()

    def generate_component_schema(self, input_bytes: bytes, schema_path: Path | str) -> None:
        generate(
            input_=input_bytes.decode("utf-8"),
            input_file_type=InputFileType.OpenAPI,
            openapi_scopes=[OpenAPIScope.Schemas],
            output=schema_path,
            output_model_type=DataModelType.PydanticV2BaseModel
            if self.use_pydantic_v2
            else DataModelType.PydanticBaseModel,
        )
        logger.info(f"generated pydantic models from schema definitions in : {schema_path}")

    def generate_service_file(
        self,
        *,
        openapi: OpenAPI,
        service_name: Optional[str] = None,
        service_path: Path | str,
        resources: list[ResourceSchema],
    ):
        services = list(
            self._build_arrest_service(openapi=openapi, service_name=service_name, resources=resources)
        )

        ServiceTemplate(services=services, destination_path=service_path).render_and_save()
        logger.info(f"generated arrest services in : {service_path}/services.py")

    def get_service_name(self, openapi: OpenAPI, service_name: Optional[str] = None) -> str:
        name = service_name or openapi.info.title or OPENAPI_DIRECTORY
        return sanitize_name(name)

    def generate_resource_file(
        self, openapi: OpenAPI, schema_path: Path | str, resource_path: Path | str
    ) -> list[ResourceSchema]:
        resources = list(self._build_arrest_resources(openapi=openapi))
        path, _ = os.path.splitext(schema_path)
        module = Path(path).stem

        ResourceTemplate(
            schema_module=module, resources=resources, destination_path=resource_path
        ).render_and_save()
        logger.info(f"generated arrest resources in : {resource_path}/resources.py")
        return resources

    def _build_arrest_service(
        self, openapi: OpenAPI, service_name: Optional[str] = None, resources: list[ResourceSchema] = None
    ) -> Generator[ServiceSchema, None, None]:
        name = service_name or self.get_service_name(openapi)

        resource_names = [res.name for res in resources]
        for idx, server in enumerate(openapi.servers):
            for idxx, url in enumerate(self._extract_url(server)):
                suffix = f"_{idx}" if idx else ""
                suffix += f"_{idxx}" if idxx else ""
                service_id = f"{name}{suffix}"
                yield ServiceSchema(
                    service_id=service_id,
                    name=name,
                    url=url,
                    description=server.description,
                    resources=resource_names,
                )

    def _extract_url(self, server: Server) -> Generator[str, None, None]:
        url, variables = server.url, server.variables
        if not variables:
            yield url
        else:
            value_list = []
            for var in variables.values():
                values = set(var.enum) if var.enum else set()
                values.add(var.default)
                value_list.append(values)
            for values in itertools.product(*value_list):
                kwargs = {k: v for k, v in zip(variables.keys(), values)}
                yield url.format(**kwargs)

    def _build_arrest_resources(self, openapi: OpenAPI) -> Generator[ResourceSchema, None, None]:
        def prefix(path: str) -> str:
            """
            ("/root/xyz/123") -> "root"
            ("/root/{id}") -> "root"
            """
            return path[1:].split("/")[0]

        for key, group in itertools.groupby(openapi.paths.keys(), key=prefix):
            routes = list(group)
            path_items = [openapi.paths.get(route) for route in routes]
            handlers: list[HandlerSchema] = []
            for route, path_item in zip(routes, path_items):
                route = route.removeprefix(f"/{key}")
                handlers.extend(self._build_handlers(route=route, path_item=path_item))

            yield ResourceSchema(name=key, route=f"/{key}", handlers=handlers)

    def _build_handlers(self, route: str, path_item: PathItem) -> list[HandlerSchema]:
        handlers = []
        for method in list(Methods):
            operation: Operation
            if operation := getattr(path_item, str(method).lower(), None):
                request_class = self.get_request_schema(operation)
                response_class = self.get_response_schema(operation)

                handlers.append(
                    HandlerSchema(route=route, method=method, request=request_class, response=response_class)
                )

        return handlers

    def get_request_schema(self, operation: Operation) -> Optional[str]:
        if not (request_body := operation.requestBody):
            logger.debug("no request body defined")
            return None

        if isinstance(request_body, Reference):
            return get_ref_schema(request_body)

        if media := request_body.content.get("application/json", None):
            return get_ref_schema(media.media_type_schema)

    def get_response_schema(self, operation: Operation) -> Optional[str]:
        if not operation.responses or not (
            success_response := operation.responses.get(str(httpx.codes.OK), None)
        ):
            logger.debug("no success (200) response defined")
            return None
        if isinstance(success_response, Reference):
            return get_ref_schema(success_response)

        if success_response.content and (media := success_response.content.get("application/json", None)):
            return get_ref_schema(media.media_type_schema)

    @classmethod
    def load_dict(cls, fmt: Format, data: IO) -> dict:
        if fmt == Format.json:
            return json.load(data)
        elif fmt in (Format.yaml, Format.yml):
            return yaml.load(data, yaml.SafeLoader)

    @classmethod
    def parse_openapi(cls, fmt: Format, data: IO) -> OpenAPI:
        spec = cls.load_dict(fmt, data)
        return OpenAPI(**spec)



from pathlib import Path
from typing import ClassVar, Optional

from pydantic import BaseModel

from arrest.openapi._base import TemplateBase


class HandlerSchema(BaseModel):
    route: str
    method: str
    request: Optional[str] = None
    response: Optional[str] = None


class ResourceSchema(BaseModel):
    name: str
    route: str
    handlers: list[HandlerSchema]


class ResourceParams(BaseModel):
    schema_module: str
    schema_imports: set[str]
    resources: list[ResourceSchema]


class ResourceTemplate(TemplateBase):
    TEMPLATE_FILEPATH: ClassVar[str] = "resources.jinja2"

    def __init__(
        self, schema_module: str, resources: list[ResourceSchema], destination_path: Path | str
    ) -> None:
        schema_imports = set()
        for resource in resources:
            for handler in resource.handlers:
                schema_imports.update((handler.request, handler.response))

        schema_imports = {imp for imp in schema_imports if imp is not None}

        super().__init__(
            source=self.TEMPLATE_FILEPATH,
            params=ResourceParams(
                schema_module=schema_module, schema_imports=schema_imports, resources=resources
            ),
            destination_path=destination_path,
        )



import posixpath
from typing import Any, Type

import orjson
from pydantic import BaseModel

from arrest._config import PYDANTIC_V2


def join_url(base_url: str, *urls: list[str]) -> str:
    path = posixpath.join(base_url, *[url.lstrip("/") for url in urls])
    if not urls[-1].endswith("/"):
        path = path.rstrip("/")
    return path


def extract_model_field(model: BaseModel, field: str) -> dict:  # pragma: no cover
    """
    reuse pydantic's own deserializer to extract single field
    as a json parsed dict
    """
    default = {}

    if PYDANTIC_V2:
        value = model.model_dump_json(include={field}, by_alias=True)
    else:
        value = model.json(include={field}, by_alias=True)
    value = orjson.loads(value)
    if not value:
        return default
    return value


def jsonify(obj: Any) -> Any:
    return orjson.loads(orjson.dumps(obj))


def validate_request_model(type_: Type[BaseModel], obj: Any) -> BaseModel:  # pragma: no cover
    if PYDANTIC_V2:
        return type_.model_validate(obj)
    return type_.parse_obj(obj)



class BaseException(Exception):
    """base exception class"""


class ArrestError(BaseException):
    """used in error situations"""

    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class ArrestHTTPException(ArrestError):
    def __init__(self, status_code: int, data: dict | str) -> None:
        self.status_code = status_code
        self.data = data


class NotFoundException(ArrestError):
    def __init__(self, message: str):
        self.message = message


class HandlerNotFound(NotFoundException):
    def __init__(self, message: str):
        super().__init__(message)


class ResourceNotFound(NotFoundException):
    def __init__(self, message: str):
        super().__init__(message)


class ConversionError(ArrestError):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)



# pylint: disable=W0707
import functools
import inspect
from typing import Any, Mapping, Optional, Type, Union, cast

import backoff
import httpx
from httpx import Headers, QueryParams
from pydantic import BaseModel, ValidationError
from pydantic.fields import FieldInfo
from typing_extensions import Unpack

from arrest._config import PYDANTIC_V2, HttpxClientInputs
from arrest.converters import compile_path
from arrest.defaults import DEFAULT_TIMEOUT, MAX_RETRIES
from arrest.exceptions import ArrestHTTPException, HandlerNotFound
from arrest.handler import HandlerKey, ResourceHandler
from arrest.http import Methods
from arrest.logging import logger
from arrest.params import Param, Params, ParamTypes
from arrest.utils import extract_model_field, join_url, jsonify, validate_request_model


class Resource:
    """
    A python class used to define a RESTful resource.

    Usage:
        ```python
        >>> from arrest import Resource

        >>> user_resource = Resource(name="user", route="/users", handlers=[("GET", "/")])
        ```

    Handlers can be provided as a list of tuple, dictionary or instances of `ResourceHandler`

    If provided as a dict or `ResourceHandler`, the keys / fields have to be set according to
    the `ResourceHandler` definiion.

    If provided as a tuple, at minimum 2 entries `(method, route)` or a maximum of 5 entries
    (method, route, request, response, callback) can be defined.
    """

    def __init__(
        self,
        name: Optional[str] = None,
        *,
        route: Optional[str],
        response_model: Optional[Type[BaseModel]] = None,
        handlers: Union[
            list[ResourceHandler],
            list[Mapping[str, Any]],
            list[tuple[Any, ...]],
        ] = None,
        client: Optional[httpx.AsyncClient] = None,
        **kwargs: Unpack[HttpxClientInputs],
    ) -> None:
        """

        Parameters:
            name:
                Unique name of the resource
            route:
                Unique route to the resource
            response_model:
                Pydantic datamodel to wrap the json response
            handlers:
                List of handlers
            client:
                An httpx.AsyncClient instance
            kwargs:
                Additional httpx.AsyncClient parameters, [see more](api.md/#httpx-client-arguments)
        """

        self._client: Optional[httpx.AsyncClient] = None
        self._httpx_args: Optional[HttpxClientInputs] = None

        self.base_url = "/"  # will be filled once bound to a service
        self.route = route

        derived_name = name if name else self.route.strip("/").split("/")[0]
        self.name = derived_name if derived_name else "root"
        self.response_model = response_model
        self.routes: dict[HandlerKey, ResourceHandler] = {}

        self.initialize_handlers(handlers=handlers)
        self.initialize_httpx(client=client, **kwargs)

    def handler(
        self,
        path: str,
    ) -> Any:
        url = join_url(self.base_url, self.route, path)

        def wrapper(func):
            @functools.wraps(func)
            @backoff.on_exception(
                backoff.expo,
                (
                    httpx.HTTPError,
                    httpx.TimeoutException,
                ),
                max_tries=MAX_RETRIES,
                jitter=backoff.full_jitter,
            )
            async def wrapped(*args, **kwargs):
                return await func(self, url, *args, **kwargs)

            if getattr(self, func.__name__, None) is None:
                setattr(self, func.__name__, wrapped)
            return wrapped

        return wrapper

    async def request(
        self,
        method: Methods,
        path: str,
        request: Union[BaseModel, Mapping[str, Any], None] = None,
        headers: Optional[Mapping[str, str]] = None,
        query: Optional[Mapping[str, str]] = None,
        **kwargs,
    ) -> Any | None:
        """
        Makes an HTTP request against a handler route

        Usage:
            ```python
            >>> user_resource.user.request(method="GET")
            ```

        Parameters:
            method:
                The HTTP method for the request
            path:
                Path to a handler specified in the resource
            request:
                A pydantic object containing the necessary fields
                to make an http request to the handler url

                **Must match the corresponding `handler.request` pydantic model**
            headers:
                A multi-dict or `httpx.Headers` containing additional
                header key-value pairs
            query:
                A multi-dict or `httpx.QueryParams` containing additional
                query-param key-value pairs
            **kwargs:
                Keyword-arguments matching the path params, if any

        Returns:
            Response:
                A JSON response in form of a list or dict
                `or`, Deserialized into the response pydantic model
                `or`, Return value of the callback fn
        """

        params: dict = {}

        if not (match := self.get_matching_handler(method=method, path=path, **kwargs)):
            logger.warning("no matching handler found for request")
            raise HandlerNotFound(message="no matching handler found for request")

        handler, url = match

        params = self.extract_request_params(
            request_type=handler.request,
            request_data=request,
            headers=headers,
            query=query,
            **kwargs,
        )

        response_type = handler.response or self.response_model or None

        response = await self._build_request(
            url=url, method=method, params=params, response_type=response_type
        )

        if handler.callback:
            try:
                if inspect.iscoroutinefunction(handler.callback):
                    callback_response = await handler.callback(response)
                else:
                    callback_response = handler.callback(response)
            except Exception:
                logger.warning("something went wrong during callback", exc_info=True)
                raise
            return callback_response

        return response

    async def get(
        self,
        path: str,
        request: Optional[BaseModel] = None,
        headers: Optional[Mapping[str, str]] = None,
        query: Optional[Mapping[str, str]] = None,
        **kwargs,
    ):
        """
        Makes a `HTTP GET` request

        see [request][arrest.resource.Resource.request]
        """
        return await self.request(
            method=Methods.GET,
            path=path,
            request=request,
            headers=headers,
            query=query,
            **kwargs,
        )

    async def post(
        self,
        path: str,
        request: Optional[BaseModel] = None,
        headers: Optional[Mapping[str, str]] = None,
        query: Optional[Mapping[str, str]] = None,
        **kwargs,
    ):
        """
        Makes a `HTTP POST` request

        see [request][arrest.resource.Resource.request]
        """
        return await self.request(
            method=Methods.POST,
            path=path,
            request=request,
            headers=headers,
            query=query,
            **kwargs,
        )

    async def put(
        self,
        path: str,
        request: Optional[BaseModel] = None,
        headers: Optional[Mapping[str, str]] = None,
        query: Optional[Mapping[str, str]] = None,
        **kwargs,
    ):
        """
        Makes a `HTTP PUT` request

        see [request][arrest.resource.Resource.request]
        """
        return await self.request(
            method=Methods.PUT,
            path=path,
            request=request,
            headers=headers,
            query=query,
            **kwargs,
        )

    async def patch(
        self,
        path: str,
        request: Optional[BaseModel] = None,
        headers: Optional[Mapping[str, str]] = None,
        query: Optional[Mapping[str, str]] = None,
        **kwargs,
    ):
        """
        Makes a `HTTP PATCH` request

        see [request][arrest.resource.Resource.request]
        """
        return await self.request(
            method=Methods.PATCH,
            path=path,
            request=request,
            headers=headers,
            query=query,
            **kwargs,
        )

    async def delete(
        self,
        path: str,
        request: Optional[BaseModel] = None,
        headers: Optional[Mapping[str, str]] = None,
        query: Optional[Mapping[str, str]] = None,
        **kwargs,
    ):
        """
        Makes a `HTTP DELETE` request

        see [request][arrest.resource.Resource.request]
        """
        return await self.request(
            method=Methods.DELETE,
            path=path,
            request=request,
            headers=headers,
            query=query,
            **kwargs,
        )

    async def head(
        self,
        path: str,
        request: Optional[BaseModel] = None,
        headers: Optional[Mapping[str, str]] = None,
        query: Optional[Mapping[str, str]] = None,
        **kwargs,
    ):
        """
        Makes a `HTTP HEAD` request

        see [request][arrest.resource.Resource.request]
        """
        return await self.request(
            method=Methods.HEAD,
            path=path,
            request=request,
            headers=headers,
            query=query,
            **kwargs,
        )

    async def options(
        self,
        path: str,
        request: Optional[BaseModel] = None,
        headers: Optional[Mapping[str, str]] = None,
        query: Optional[Mapping[str, str]] = None,
        **kwargs,
    ):
        """
        Makes a `HTTP OPTIONS` request

        see [request][arrest.resource.Resource.request]
        """
        return await self.request(
            method=Methods.OPTIONS,
            path=path,
            request=request,
            headers=headers,
            query=query,
            **kwargs,
        )

    def extract_request_params(
        self,
        request_type: Type[BaseModel] | None,
        request_data: BaseModel | Mapping[str, Any] | None,
        headers: Mapping[str, str] | None = None,
        query: Mapping[str, Any] | None = None,
    ) -> Params:
        """
        extracts `header`, `body` and `query` params from the pydantic request model

        Parameters:
            request_type:
                a pydantic class for holding the request data
            request_data:
                instance of the above containing the data
            kwargs:
                optional keyword-arguments containing query parameters
        Returns:
            a dictionary containing `header`, `body`, `query` params in separate dicts
        """

        header_params = headers or {}
        query_params = query or {}
        body_params = {}

        if request_type:
            # perform type validation on `request_data`
            request_data = validate_request_model(type_=request_type, obj=request_data)

        if isinstance(request_data, BaseModel):
            # extract pydantic fields into `Query`, `Body` and `Header`
            model_fields: dict = request_data.model_fields if PYDANTIC_V2 else request_data.__fields__

            for field_name, field in model_fields.items():
                field_info = field if PYDANTIC_V2 else field.field_info
                field_info = cast(Param, field_info)
                if not hasattr(field_info, "_param_type") and isinstance(field_info, FieldInfo):
                    body_params |= extract_model_field(request_data, field_name)
                elif field_info._param_type == ParamTypes.query:
                    query_params |= extract_model_field(request_data, field_name)
                elif field_info._param_type == ParamTypes.header:
                    header_params |= extract_model_field(request_data, field_name)
                elif field_info._param_type == ParamTypes.body:
                    body_params |= extract_model_field(request_data, field_name)

        else:
            body_params = request_data

        return Params(
            header=Headers(header_params),
            query=QueryParams(query_params),
            body=jsonify(body_params),
        )

    @backoff.on_exception(
        backoff.expo,
        (
            httpx.HTTPError,
            httpx.TimeoutException,
            ArrestHTTPException,
        ),
        max_tries=MAX_RETRIES,
        jitter=backoff.full_jitter,
    )
    async def _build_request(
        self,
        url: str,
        method: Methods,
        params: Params,
        response_type: Optional[Type[BaseModel]],
    ) -> Any:
        """
        (private) makes the actual http call using httpx

        Parameters
        ----------

        url : str
            the complete url of the endpoint
        method : Methods
            one of the available http methods
        params : dict
            dict containing `header`, `query` and `body` params
        response_type : Optional[Type[BaseModel]]
            optional response_type to deserialize the json response to

        Returns
        -------

        Any
            a json object or a parsed pydantic object
        """

        try:
            if self._client:
                response = await self._make_request(
                    client=self._client, url=url, method=method, params=params
                )
            else:
                async with httpx.AsyncClient(base_url=self.base_url, **self._httpx_args) as client:
                    response = await self._make_request(client=client, url=url, method=method, params=params)

            status_code = response.status_code
            logger.debug(f"{method!s} {url} returned with status code {status_code!s}")
            response.raise_for_status()
            response_body = response.json()

            # parse response to pydantic model
            parsed_response = response_body
            if response_type:
                response_type
                if isinstance(response_body, list):
                    parsed_response = [response_type(**item) for item in response_body]
                elif isinstance(response_body, dict):
                    parsed_response = response_type(**response_body)
                else:
                    raise ValueError(f"could not parse response to pydantic model {response_type.__name__}")

            return parsed_response

        # exception handling
        except httpx.HTTPStatusError as exc:
            err_response_body = exc.response.json()
            raise ArrestHTTPException(status_code=exc.response.status_code, data=err_response_body) from exc

        except httpx.TimeoutException:
            raise ArrestHTTPException(
                status_code=httpx.codes.INTERNAL_SERVER_ERROR,
                data="request timed out",
            )

        except httpx.RequestError:
            raise ArrestHTTPException(
                status_code=httpx.codes.INTERNAL_SERVER_ERROR,
                data="error occured while making request",
            )

    async def _make_request(
        self, client: httpx.AsyncClient, url: str, method: Methods, params: Params
    ) -> httpx.Response:
        header_params, query_params, body_params = (
            params.header,
            params.query,
            params.body,
        )
        match method:
            case Methods.GET:
                response = await client.get(url=url, params=query_params, headers=header_params)
            case Methods.POST:
                response = await client.post(
                    url=url,
                    params=query_params,
                    headers=header_params,
                    json=body_params,
                )
            case Methods.PUT:
                response = await client.put(
                    url=url,
                    params=query_params,
                    headers=header_params,
                    json=body_params,
                )
            case Methods.PATCH:
                response = await client.patch(
                    url=url,
                    params=query_params,
                    headers=header_params,
                    json=body_params,
                )
            case Methods.DELETE:
                response = await client.delete(
                    url=url,
                    params=query_params,
                    headers=header_params,
                )
            case Methods.HEAD:
                response = await client.head(url=url, params=query_params, headers=header_params)

            case Methods.OPTIONS:
                response = await client.options(url=url, params=query_params, headers=header_params)

        return response

    def get_matching_handler(
        self, method: Methods, path: str, **kwargs
    ) -> tuple[ResourceHandler, str] | None:
        for handler in self.routes.values():
            parsed_path = handler.parse_path(method=method, path=path, **kwargs)
            if parsed_path is not None:
                url = join_url(self.route, parsed_path)
                return handler, url

    def _bind_handler(self, base_url: str | None = None, *, handler: ResourceHandler) -> None:
        """
        compose a fully-qualified url by joining base service url, resource url
        and handler url,
        and map that to the handler object.
        Note: If there are duplicate handlers or any combination resulting in
        duplicate fq-url, the later will take precendence.
        """

        base_url = base_url or self.base_url
        handler.path_regex, handler.path_format, handler.param_types = compile_path(handler.route)

        self.routes[HandlerKey(*(handler.method, handler.path_format))] = handler

    def initialize_handlers(
        self,
        base_url: str | None = None,
        handlers: list[ResourceHandler] | list[Mapping[str, Any]] | list[tuple[Any, ...]] = None,
    ) -> None:
        """
        specifically used to inject `base_url` from a Service class to
        downstream Resources.
        Should not be used separately (unless you want to break things)
        """
        if self.routes:
            handlers = list(self.routes.values())

        if not handlers:
            return

        for _handler in handlers:
            try:
                if isinstance(_handler, dict):
                    self._bind_handler(base_url=base_url, handler=ResourceHandler(**_handler))
                elif isinstance(_handler, tuple):
                    if len(_handler) < 2:
                        raise ValueError("Too few arguments to unpack. Expected atleast 2")
                    if len(_handler) > 5:
                        raise ValueError(f"Too many arguments to unpack. Expected 5, got {len(_handler)}")

                    method, route, rest = (
                        _handler[0],
                        _handler[1],
                        _handler[2:],
                    )

                    self._bind_handler(
                        base_url=base_url,
                        handler=ResourceHandler(
                            method=method,
                            route=route,
                            request=len(rest) >= 1 and rest[0] or None,
                            response=len(rest) >= 2 and rest[1] or None,
                            callback=len(rest) >= 3 and rest[2] or None,
                        ),
                    )
                elif isinstance(_handler, BaseModel):
                    self._bind_handler(base_url=base_url, handler=_handler)
                else:
                    raise ValueError("invalid handler type specified")
            except ValidationError:
                raise ValueError("cannot initialize handler signature")

    def initialize_httpx(
        self, client: Optional[httpx.AsyncClient], **kwargs: Unpack[HttpxClientInputs]
    ) -> None:
        if client:
            self._client = client
        elif self._httpx_args:
            # already initialized
            self._httpx_args |= kwargs
        else:
            timeout = kwargs.get("timeout")
            if not timeout:
                kwargs["timeout"] = httpx.Timeout(timeout=DEFAULT_TIMEOUT, connect=DEFAULT_TIMEOUT)

            if isinstance(timeout, int):
                kwargs["timeout"] = httpx.Timeout(timeout=timeout, connect=timeout)

            self._httpx_args = HttpxClientInputs(**kwargs)



import enum


class StrEnum(str, enum.Enum):
    def __repr__(self) -> str:
        return self.value

    def __str__(self) -> str:
        return self.__repr__()


class Exit(int, enum.Enum):
    """Exit reasons."""

    OK = 0
    ERROR = 1
    KeyboardInterrupt = 2



"""
Set of pydantic wrappers around openapi dict
supports openapi >= v3.0
"""
from typing import Optional, Union

from pydantic import BaseModel, ConfigDict, Extra, Field

from arrest._config import PYDANTIC_V2


class Base(BaseModel):
    if PYDANTIC_V2:
        model_config = ConfigDict(extra="allow", populate_by_name=True)

    else:

        class Config:
            extra = Extra.allow
            allow_population_by_field_name = True


class Info(Base):
    title: str
    version: str


class ServerVariable(Base):
    enum: Optional[list[str]] = None
    default: str


class Server(Base):
    url: str
    description: Optional[str] = None
    variables: Optional[dict[str, ServerVariable]] = None


class Reference(Base):
    ref: str = Field(alias="$ref")


class MediaType(BaseModel):
    media_type_schema: Optional[Union[Reference, dict]] = Field(default=None, alias="schema")


class RequestBody(BaseModel):
    content: dict[str, MediaType]


class Response(Base):
    content: Optional[dict[str, MediaType]] = None


class Operation(Base):
    responses: Optional[dict[str, Union[Reference, Response]]] = None
    requestBody: Optional[Union[Reference, RequestBody]] = None


class PathItem(Base):
    ref: Optional[str] = Field(default=None, alias="$ref")
    get: Optional[Operation] = None
    post: Optional[Operation] = None
    put: Optional[Operation] = None
    patch: Optional[Operation] = None
    delete: Optional[Operation] = None
    head: Optional[Operation] = None
    options: Optional[Operation] = None


class OpenAPI(Base):
    info: Info
    servers: list[Server] = [Server(url="/")]
    paths: dict[str, PathItem]



import sys
from argparse import Namespace
from pathlib import Path
from typing import Optional, Sequence

import argcomplete

from arrest.common import Exit
from arrest.logging import logger

try:
    import datamodel_code_generator
    import jinja2
except ImportError:
    logger.warning("Dependencies missing. Please install extra dependencies by `pip install arrest[openapi]`")
    sys.exit(Exit.ERROR)

from arrest.cli.arguments import arg_parser
from arrest.openapi.parser import OpenAPIGenerator


def main(args: Optional[Sequence[str]] = None):
    namespace = Namespace()
    argcomplete.autocomplete(arg_parser)

    if args is None:  # pragma: no cover
        args = sys.argv[1:]

    _, unknown_args = arg_parser.parse_known_args(args, namespace=namespace)

    if unknown_args:
        print("unrecognized arguments: {}".format(" ".join(unknown_args)))
        arg_parser.print_help(file=sys.stdout)
        return Exit.ERROR

    if not args:
        arg_parser.print_help(file=sys.stdout)
        return Exit.ERROR

    if not namespace.url:
        print("Missing `--url`. An http or file url needs to be specified", file=sys.stdout)
        arg_parser.print_help(file=sys.stdout)
        return Exit.ERROR

    if not (output := namespace.output):
        output = Path().resolve(namespace.output)
        print(f"No output path specified. Using current directory {output!s}", file=sys.stdout)

    use_pydantic_v2 = namespace.pydantic == "v2"

    try:
        generator = OpenAPIGenerator(
            url=namespace.url,
            output_path=output,
            dir_name=namespace.dir,
            use_pydantic_v2=use_pydantic_v2,
        )

        generator.generate_schema()
        print("Files generated successfully", file=sys.stdout)
        return Exit.OK

    except Exception:
        import traceback

        print(traceback.format_exc(), file=sys.stderr)
        return Exit.ERROR


if __name__ == "__main__":
    sys.exit(main())



from enum import Enum
from typing import Any

import httpx
from pydantic.fields import FieldInfo

try:  # pragma: no cover
    from pydantic_core import PydanticUndefined
except ImportError:
    from pydantic.fields import Undefined as PydanticUndefined
from typing_extensions import Unpack


class ParamTypes(Enum):
    query = "query"
    header = "header"
    body = "Body"


class Param(FieldInfo):
    _param_type: ParamTypes

    def __init__(self, default: Any = PydanticUndefined, **kwargs: Unpack[Any]) -> None:
        kwargs = dict(default=default, **kwargs)
        super().__init__(**kwargs)


class Query(Param):
    _param_type = ParamTypes.query

    def __init__(self, default: Any = PydanticUndefined, **kwargs: Unpack[Any]) -> None:
        super().__init__(default, **kwargs)


class Header(Param):
    _param_type = ParamTypes.header

    def __init__(self, default: Any = PydanticUndefined, **kwargs: Unpack[Any]) -> None:
        super().__init__(default, **kwargs)


class Body(Param):
    _param_type = ParamTypes.body

    def __init__(self, default: Any = PydanticUndefined, **kwargs: Unpack[Any]) -> None:
        super().__init__(default, **kwargs)


class Params:
    def __init__(self, *, header: httpx.Headers, query: httpx.QueryParams, body: Any) -> None:
        self.header = header
        self.query = query
        self.body = body
